////////////////////////////////////////////////////////////////////
//
// File: 		setfv.cpp
// Project: 	setfv
//
// Desc:		Sample app on changin version info
//
// Revisions: 	Created 08/01/99
//				Modified 29/03/99: Changed some pointer alignments
//
// Copyright(C) 1999, Tomas Restrepo. All rights reserved
//
///////////////////////////////////////////////////////////////////

// UNICODE Stuff
// Although not obvious
// the program MUST be compiled with unicode
// I wrote it using generic mappings
// to allow debugging under Win9x
// because the version strings get translated to ansi in
// memory. On NTm, however, they are always unicode
#define UNICODE
#ifdef UNICODE
#define _UNICODE
#endif
#pragma comment(lib, "version.lib")
#pragma comment(lib, "user32.lib")

#define STRICT
#define WIN32_LEAN_MEAN
#include <windows.h>
#include <stdio.h>
#include <tchar.h>

// a VAR struct represents the first
// part of all variable-length
// structures in the version resource
#pragma pack (4)
typedef struct _tagVAR
{
	WORD	wLength;
	WORD	wValueLength;
	WORD	wType;
	TCHAR	szKey[1];	// ansi in Win95, Unicode on NT
} VAR, *LPVAR;
#pragma pack ()

// some useful Macros
// word inverter
#define INVERT_WORDS(dword) (((dword & 0xFFFF0000) >> 16) | ((dword & 0x0000FFFF) << 16))
// byte length of string, with terminating '\0'
#define STRLEN(lpstring) ((_tcslen(lpstring)+1) * sizeof(TCHAR))
// next mem-address align on DWORD boundaries
#define ALIGN_DWORD(x) ( (x & 0xFFFFFFFC)?(x & 0xFFFFFFFC) + 4 : x )
// returns next VAR struct in array
#define NEXT_VAR(lpVar) (LPVAR)ALIGN_DWORD((DWORD)lpVar + lpVar->wLength)


BOOL CALLBACK MyLangEnumProc ( HINSTANCE hModule, 
							   LPCTSTR lpszType, 
							   LPCTSTR lpszName, 
							   WORD wIDLanguage,
							   LONG lParam )
{
	WORD	*plangid = (WORD*)lParam;
	*plangid = wIDLanguage;
	return FALSE;	// end enumeration since we're only interested 
					// in the first resource
}


WORD GetVersionLangId ( LPTSTR lpszFile )
{
	WORD		langid = MAKELANGID ( LANG_NEUTRAL, SUBLANG_NEUTRAL ); // default
	HINSTANCE	hInst;

	// load the library
	hInst = LoadLibraryEx ( lpszFile, NULL, LOAD_LIBRARY_AS_DATAFILE );
	if ( hInst == NULL )
		return langid;

	EnumResourceLanguages ( hInst, RT_VERSION, MAKEINTRESOURCE(VS_VERSION_INFO), 
							MyLangEnumProc, (LONG)&langid );

	
	FreeLibrary ( hInst );
	return langid;
}


// GetValue ()
// This is a helper function that will
// return a LPBYTE, pointing to the value
// or child member of the VAR struct
// passed as a parameter.The value member will be
// aligned on DWORD boundaries
LPBYTE GetValue ( LPVAR lpVar )
{
	DWORD	addr = 0;

	addr = (DWORD)lpVar + sizeof(lpVar) + STRLEN( lpVar->szKey );
	addr += addr % 4;
	return (LPBYTE) addr;
} 

// UpdateVersionString()
// Looks up the "FileVersion"
// section on the resource, and 
// updates the string
BOOL UpdateVersionString ( LPBYTE lpData,
						   LPTSTR lpszVersion )
{
	LPVAR		lpTable = NULL; 
	LPVAR		lpString = NULL;
	DWORD		dwBuffSize = 0; // size of buffer
	LPBYTE		lpBuffer = NULL; // temporary space
	LPBYTE		lpOffset = NULL; // offset from which to copy to the buffer
	UINT		uLen;
	int			nStrDif = 0; 

	if ( !VerQueryValue ( (LPVOID)lpData, _T("\\StringFileInfo"), 
						  (LPVOID*)&lpTable, &uLen ) )
		return FALSE;

	// now lpTable points to the first child of a StringFileInfo
	// struct, which is a StringTable struct. The childs of this struct
	// are String structs, each identified by a specific value 
	// in the szKey member. Search them until we find "FileVersion"
	lpString = (LPVAR)GetValue ( lpTable );

	dwBuffSize = (DWORD)NEXT_VAR ( lpTable );
	while ( ((DWORD)lpString) < dwBuffSize)
	{
		if ( _tcscmp ( lpString->szKey, _T("FileVersion") ) != 0 )
			lpString = NEXT_VAR ( lpString );
		else
		{
			// we found the string
			// now, before continuing, copy the original
			// resource to a temporary buffer. After updating 
			// the string, we'll copy it back
			//
			// calc the buffer size needed
			dwBuffSize = ((LPVAR)lpData)->wLength - ((LPBYTE)NEXT_VAR(lpString)-lpData);
			
			lpBuffer = (LPBYTE) malloc ( dwBuffSize );
			if ( lpBuffer == NULL ) return FALSE;
			memcpy ( lpBuffer, NEXT_VAR(lpString), dwBuffSize );
			
			// NOTE:
			// Docs say the strings are unicode, and that's how
			// they are saved in the file. However, when loaded 
			// on  memory on 9x, they get translated first to ansi
			// and appear so, followed by those in unicode.
			// however, we don't care, since UpdateResource()
			// and friends are NT only.
			// update the string, but save first the
			// difference in bytes between the old and new string
			// so we can update the resource structures after that.
			nStrDif = STRLEN(lpszVersion) - STRLEN((LPTSTR)GetValue(lpString));
			memcpy ( GetValue ( lpString ), lpszVersion, STRLEN(lpszVersion));
			// wValueLength should be in words in this case
			lpString->wValueLength += nStrDif / sizeof(WORD);
			// all other updates are in bytes
			lpString->wLength += nStrDif; 
			lpTable->wLength += nStrDif; 
			((LPVAR)((LPBYTE)lpTable - 36))->wLength += nStrDif; // hack
			((LPVAR)lpData)->wLength += nStrDif; 
			
			// copy buffer back, and free the memory used;
			// notice that by this time the value returned
			// by NEXT_VAR is probably different than before
			memcpy ( (LPBYTE)ALIGN_DWORD((DWORD)NEXT_VAR(lpString)), lpBuffer, dwBuffSize );
			free ( lpBuffer );
			
			return TRUE;
		}
	}

	// didn't find it, return
	return FALSE;

}

BOOL SetFileVersion ( LPTSTR lpszFile, 
					  WORD wMajor, 
					  WORD wMinor,
					  WORD wBuild,
					  WORD wFinal )
{
	HANDLE				hFile = NULL;
	LPBYTE				lpData = NULL;
	LPBYTE				pos = NULL; // for strings manipulation
	DWORD				dwSize, dwHandle;//, dwLangInfo;
	WORD				langid;
	TCHAR				version[30];
	UINT				uLen = 0;
	VS_FIXEDFILEINFO	*lpffi = NULL;

	
	// get langid of resource, used later for saving
	langid = GetVersionLangId ( lpszFile );
	// get size of buffer
	dwSize = GetFileVersionInfoSize ( lpszFile, &dwHandle );
	if ( dwSize == 0 ) // no version resource, get outta here
		goto _FAILED;
	// add some space to the buffer,
	// so that we can modify the strings later
	// I'm too lazy to calculate the space needed
	// before hand, so we'll use a fixed size
	dwSize += 100;
	lpData = (LPBYTE)malloc ( dwSize );
	if ( lpData == NULL ) // out of memory
		goto _FAILED;
	// now get the data
	if ( !GetFileVersionInfo ( lpszFile, 0, dwSize, (LPVOID)lpData ) )
		goto _FAILED;
	// get the VS_FIXEDFILEINFO address
	if ( VerQueryValue ( (LPVOID)lpData, _T("\\"), (LPVOID*)&lpffi, &uLen ) == 0 || uLen == 0)
		goto _FAILED;

	// change version in the most important fields
	lpffi->dwFileVersionMS = MAKELONG ( wMinor, wMajor );
	lpffi->dwFileVersionLS = MAKELONG ( wFinal, wBuild );


	// now comes the tough part. We've also have to update one string
	// that has the build number stored on it.
	// update the version string
	wsprintf ( version, _T("%hu.%hu.%hu.%hu"), wMajor, wMinor, wBuild, wFinal );
	UpdateVersionString ( lpData, version );

	// begin update
	hFile = BeginUpdateResource ( lpszFile, FALSE );
	if ( hFile == NULL )
		goto _FAILED;

	if ( !UpdateResource ( hFile, RT_VERSION, MAKEINTRESOURCE(VS_VERSION_INFO), 
						   langid, (LPVOID)lpData, dwSize ) )
		goto _FAILED;
	

	if ( EndUpdateResource ( hFile, FALSE ) == 0 )
		goto _FAILED;

	free ( lpData );
	return TRUE;

_FAILED:
	if ( lpData != NULL )
		free ( lpData );
	return FALSE;
}

int _tmain ( int argc, TCHAR *argv[] )
{
	WORD	wMajor = 0, 
			wMinor = 0,
			wBuild = 0,
			wFinal = 0;

	if ( argc != 3 )
	{
		puts ( "Usage: setfv <file> <version>" );
		puts ( "version is a string in the form Major.Minor.Build.Final" );
		return 0;
	}
	
	// parse file version in Major, Minor and Build Numbers
	_stscanf ( argv[2], _T("%hu.%hu.%hu.%hu"), &wMajor, &wMinor, &wBuild, &wFinal );

	if ( !SetFileVersion ( argv[1], wMajor, wMinor, wBuild, wFinal ) )
		printf ( "Failed to set file version, error: %lu", GetLastError () );
	else
		puts ( "File version change succeeded" );
		
	return 0;
} // main